package org.jboss.seam.entityconverter;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.Converter;
import javax.faces.convert.ConverterException;
import javax.faces.el.ValueBinding;

import org.jboss.seam.core.FacesMessages;
import org.jboss.seam.core.Messages;

public abstract class AbstractEntityConverter<E> implements Converter {
	
	private ValueBinding lookup; 

	private static Map<String, String> defaultMessages;

	static {
		defaultMessages = new HashMap<String, String>();
		defaultMessages.put(
				"org.jboss.seam.entityconverter.entityConverter.idNotFound",
				"Cannot find a method or field annotated with @Id");
		defaultMessages.put(
				"org.jboss.seam.entityconverter.entityConverter.notAnEntity",
				"Referenced object is not annotated with @Entity");
		defaultMessages
				.put(
						"org.jboss.seam.entityconverter.entityConverter.entityClassNotFound",
						"Cannot find class referenced in entityClass attribute");
		defaultMessages.put(
				"org.jboss.seam.entityconverter.entityConverter.classNotFound",
				"Cannot determine class of entity to load");
		defaultMessages
				.put(
						"org.jboss.seam.entityconverter.entityConverter.emAccessError",
						"Cannot access entityManager; Either use the default name (entityManager) for your entity manager or specify it in components.xml");
		defaultMessages.put(
				"org.jboss.seam.entityconverter.entityConverter.invalidSelection",
				"Invalid selection.  Selected item cannot be loaded from persistence context");
	}

	/**
	 * Utitliy method to get a FacesMessage for ConverterExceptions
	 * 
	 * @param message
	 *            The message to create
	 * @return The FacesMessage Object
	 */
	protected FacesMessage getMessage(String message) {
		if (!message.equals( Messages.instance().get(message))) {
			return FacesMessages.createFacesMessage(FacesMessage.SEVERITY_FATAL,
					(String) Messages.instance().get(message));
		} else {
			return FacesMessages.createFacesMessage(FacesMessage.SEVERITY_FATAL,
				(String) defaultMessages.get(message));
		}
	}
	
	public static final String EM_ACCESS_ERROR = "org.jboss.seam.entityconverter.entityConverter.emAccessError";
	public static final String INVAlID_SELECTION = "org.jboss.seam.entityconverter.entityConverter.invalidSelection";

	/**
	 * This method should not be overridden (non-Javadoc)
	 * 
	 * @see org.jboss.seam.entityconverter.SelectItemsConverter#convertToString(javax.faces.context.FacesContext,
	 *      javax.faces.component.UIComponent, java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	public String getAsString(FacesContext faces, UIComponent c,
			Object obj) {
		if (obj == null) {
			return null;
		}
		// For some reason MyFaces will sometimes try to reconvert an already
		// converted objct so
		// pass through
		if (obj instanceof String)
			return (String) obj;
		return getIdAsString((E) obj);
	}

	/**
	 * This method should not be overridden (non-Javadoc)
	 * 
	 * @see org.jboss.seam.entityconverter.SelectItemsConverter#convertToObject(javax.faces.context.FacesContext,
	 *      javax.faces.component.UIComponent, java.lang.String)
	 */
	public Object getAsObject(FacesContext facesContext, UIComponent c,
			String value) {
		if (value == null) {
			return null;
		}
		Class clazz = getEntityClass(c, facesContext);
		Object entity = loadEntityFromPersistenceContext(clazz, getIdAsObject(value,
				clazz));
		return entity;
	}

	/**
	 * Override this method to provide a custom conversion from Entity to String
	 * It should return a unique, unchanging, reference to the entity
	 * 
	 * @param entity
	 * @return
	 */
	protected abstract String getIdAsString(E entity);

	/**
	 * The class of the entity to load This method can be overriden in a simple
	 * fashion e.g.
	 * 
	 * <code>return MyEntity.class;</code>
	 * 
	 * @param cmp
	 *            The JSF component which the conversion is for
	 * @param facesContext
	 * @return The class of the entity to load,
	 * @throws ConverterException
	 *             If unable to determine the EntityClass
	 */
	protected Class getClazz() {
         Type type = getClass().getGenericSuperclass();
         if (type instanceof ParameterizedType)
         {
            ParameterizedType paramType = (ParameterizedType) type;
            return (Class<E>) paramType.getActualTypeArguments()[0];
         }
         else
         {
            throw new IllegalArgumentException("Could not guess entity class by reflection");
         }
	}
	
	/**
	 * The EntityConverter attempts to guess the entity's class
	 * If the EntityConverter is extended using generics, or is invoked
	 * through si:selectitems or si:convertEntity this should work.
	 * 
	 * Otherwise override getClazz() to specify the class
	 * @param cmp
	 * @param facesContext
	 * @return
	 */
	protected Class getEntityClass(UIComponent cmp, FacesContext facesContext) {
		return getClazz();
	}

	/**
	 * Retrieve the Entity from the PersistenceContext The EnityManagerFactory
	 * specified using persistenceUnitJndiName is used
	 * 
	 * @param clazz
	 *            The class of the entity to load
	 * @param id
	 *            The id of the entity to load
	 * @return The entity
	 * @throws ConverterException
	 *             if retrieval of the entity failed
	 */
	@SuppressWarnings("unchecked")
	private Object loadEntityFromPersistenceContext(Class clazz, Object id) {
		Object entity = null;
		try {
			entity = _getLookup(FacesContext.getCurrentInstance()).find(clazz, id);
		} catch (Exception e) {
			throw new ConverterException(getMessage(EM_ACCESS_ERROR), e);
		}
		if (entity == null) {
			throw new ConverterException(
					getMessage(INVAlID_SELECTION));
		} else {
			return entity;
		}

	}

	/**
	 * This method should be overriden in a custom EntityConverter. Given the id
	 * (as a String) and the class of the entity, it should return the Id of the
	 * object in its native type
	 * 
	 * @param id
	 *            The Id of the entity
	 * @param clazz
	 *            The class of the entity
	 * @return The id of the entity as a correctly typed object
	 */
	protected abstract Object getIdAsObject(String id, Class clazz);
	
	public ValueBinding getLookup() {
		return lookup;
	}
	
	public void setLookup(ValueBinding lookup) {
		this.lookup = lookup;
	}
	
	private Lookup _getLookup(FacesContext facesContext) {
		if (lookup != null) {
			return (Lookup) lookup.getValue(facesContext);
		} else {
			return Lookup.instance();
		}
	}

}