package org.jboss.seam.entityconverter;


import java.lang.reflect.Field;
import java.lang.reflect.Method;

import javax.faces.component.StateHolder;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.ConverterException;
import javax.faces.el.ValueBinding;
import javax.persistence.Entity;
import javax.persistence.Id;

import org.apache.commons.beanutils.ConvertUtils;
import org.jboss.seam.ui.NoSelectionConverter;

/**
 * This implementation of the EntityConverter is suitable for
 * any Entity which has an primitive or String id 
 * i.e. int  etc, double etc., char, String
 * 
 *
 */
public class EntityConverter extends AbstractEntityConverter implements StateHolder{
	
	public EntityConverter() {
	}
	
	public static final String ID_NOT_FOUND="org.jboss.seam.entityconverter.entityConverter.idNotFound";
	public static final String NOT_AN_ENTITY="org.jboss.seam.entityconverter.entityConverter.notAnEntity";
	public static final String ENTITY_CLASS_NOT_FOUND="org.jboss.seam.entityconverter.entityConverter.entityClassNotFound";
	public static final String CLASS_NOT_FOUND = "org.jboss.seam.entityconverter.entityConverter.classNotFound";
	
	/**
	 * @param entity The entity to use
	 * @return The ID of the entity as a string
	 */
	protected String getIdAsString(Object entity) {
		Field idField = findIdField(entity);
		Method idMethod = findIdProperty(entity);
		if (idField != null) {
			try {
				Object id = idField.get(entity);
				if (id == null) {
					return NoSelectionConverter.NO_SELECTION_VALUE;
				} else {
					return id.toString();
				}
			} catch (Exception e) {
				throw new ConverterException(getIdNotFoundMessage(entity), e);
			}
		} else if (idMethod != null) {
			Object[] params = {};
			try {
				Object id = idMethod.invoke(entity, params);
				if (id == null) {
					return NoSelectionConverter.NO_SELECTION_VALUE;
				} else {
					return id.toString();
				}
			} catch (Exception e) {
				throw new ConverterException(
						getIdNotFoundMessage(entity), e);
			}
		} else {
			throw new ConverterException(
					getMessage(getIdNotFoundMessage(entity)));
		}
	}

	private String getIdNotFoundMessage(Object entity) {
		// Check that the class passed is an entity
		String message = getMessage(ID_NOT_FOUND) + " " + entity.getClass().getName();
		if (!(entity.getClass().isAnnotationPresent(Entity.class))) {
			message += " (additionally " + getMessage(NOT_AN_ENTITY) + ")";
		}
		return message;
	}
	
	
	/**
	 * Utility method to find Id field on an entity
	 * @param entity The entity to find the id field on
	 * @return The Id field, null if not found
	 * @throws ConverterException if the Class is not an entity
	 */
	protected Field findIdField(Object entity) {
		Class clazz = entity.getClass();
		return findIdField(clazz);
	}

	/**
	 * Find Id field on an entity
	 * @param clazz The clazz of the entity to find the id field on
	 * @return The Id field, null if not found
	 * @throws ConverterException if the Class is not an entity
	 */
	@SuppressWarnings("unchecked")
	protected Field findIdField(Class clazz) {
		for (Class currClazz = clazz; !currClazz.equals(Object.class); currClazz = currClazz.getSuperclass()) {
			// Iterate over the fields and find the Id field
			for (Field f : currClazz.getDeclaredFields()) {
				if (!(f.isAccessible()))
					f.setAccessible(true);
				if (f.isAnnotationPresent(Id.class)) {
					return f;
				}
			}
		}
		return null;
	}
	
	/**
	 * Find the Id property (getter method) on an entity
	 * @param entity The entity to find the id property
	 * @return The Id property, null if not found
	 * @throws ConverterException if the Class is not an entity
	 */
	protected Method findIdProperty(Object entity) {
		return findIdMethod(entity.getClass());
	}

	/**
	 * Find the Id property (getter method) on an entity
	 * @param clazz The entity to find the id property on
	 * @return The Id property, null if not found
	 * @throws ConverterException if the Class is not an entity
	 */
	@SuppressWarnings("unchecked") 
	protected Method findIdMethod(Class clazz) {
		for (Class currClazz = clazz; !currClazz.equals(Object.class); currClazz = currClazz.getSuperclass()) { 
			// Iterate over the fields and find the Id field
			for (Method m : currClazz.getDeclaredMethods()) {
				if (!(m.isAccessible()))
					m.setAccessible(true);
				if (m.isAnnotationPresent(Id.class)) {
					return m;
				}
			}
		}
		return null;
	}

	protected Object getIdAsObject(String id, Class clazz) {
		Field idField = findIdField(clazz);
		Method idMethod = findIdMethod(clazz);
		Class idClass;
		if (idField != null) {
			idClass = idField.getType();
		} else if (idMethod != null) {
			idClass = idMethod.getReturnType();
		} else {
			throw new ConverterException(
					getMessage("Unable to read id from entity"));
		}
		return ConvertUtils.convert(id, idClass);
	}
	
	protected Class getEntityClass(UIComponent cmp, FacesContext facesContext) {
		// If the class parameter has been specified use it
		if (entityClazz != null && !entityClazz.equals("") ) {
			try {
				return Class.forName(entityClazz);
			} catch (ClassNotFoundException e) {
				throw new ConverterException(getMessage(ENTITY_CLASS_NOT_FOUND + " " + entityClazz).getDetail(), e);
			}
		} else {
			// Otherwise try to find out by value binding
			ValueBinding vb = cmp.getValueBinding("value");
			if (vb == null) {
				throw new ConverterException(
						getMessage(CLASS_NOT_FOUND));
			}
			return vb.getType(facesContext);
		}
	}
	
	public Object saveState(FacesContext arg0) {
		Object[] values = new Object[1];
		values[0] = entityClazz;
		return values;
	}

	public void restoreState(FacesContext arg0, Object state) {
		Object[] values = (Object[])state;
		entityClazz = (String) values[0];
	}
	
	private String entityClazz;
	
	public String getEntityClass() {
		return entityClazz;
	}

	public void setEntityClass(String clazz) {
		this.entityClazz = clazz;
	}

	/** 
	 * Use the more powerful version getEntityClass(cmp, ctx)
	 */
	@Override
	protected Class getClazz() {
		return null;
	}

	private boolean _transient;
	
	public boolean isTransient() {
		return _transient;
	}

	public void setTransient(boolean newTransientValue) {
		this._transient = newTransientValue;
	}
	
}
